@@ -11,6 +11,7 @@ class Decorators:
         self._ws_channels = {}
         self._maintern_funcs = {}
         self._mount_option = {}
+        self._testpasses = {}
         self._decos = {
             "bearer_handler": self.default_bearer_handler
         }
@@ -24,8 +25,9 @@ class Decorators:
         self._binds_request = [None] * 4
 
     # function param saver ------------------------------------------
-    def get_func_id (self,  func):
-        return ("ns" in self._mount_option and self._mount_option ["ns"] + "." or "") + func.__name__
+    def get_func_id (self, func, mount_option = None):
+        mount_option = mount_option or self._mount_option
+        return ("ns" in mount_option and mount_option ["ns"] + "." or "") + func.__name__
 
     def save_function_spec (self, func):
         # save original function spec for preventing distortion by decorating wrapper
@@ -35,10 +37,11 @@ class Decorators:
             # save origin spec
             self._function_specs [func_id] = inspect.getfullargspec (func)
             self._current_function_specs [func_id] = None
+        return func_id
 
-    def get_function_spec (self, func):
-        # called by websocet_handler
-        func_id = self.get_func_id (func)
+    def get_function_spec (self, func, mount_option = None):
+        # called by websocet_handler with mount_option
+        func_id = self.get_func_id (func, mount_option)
         return self._function_specs.get (func_id)
 
     # app life cycling -------------------------------------------
@@ -94,10 +97,12 @@ class Decorators:
 
     def testpass_required (self, testfunc):
         def decorator(f):
-            self.save_function_spec (f)
+            func_id = self.save_function_spec (f)
+            self._testpasses [func_id] = testfunc
             self.set_auth_flag (f, ('testpass', testfunc.__name__))
             @wraps(f)
             def wrapper (was, *args, **kwargs):
+                testfunc = self._testpasses [func_id]
                 response = testfunc (was)
                 if response is False:
                     return was.response ("403 Permission Denied")
@@ -110,34 +115,78 @@ class Decorators:
     # parameter helpers ------------------------------------------------
     RX_EMAIL = re.compile (r"[a-z0-9][-.a-z0-9]*@[-a-z0-9]+\.[-.a-z0-9]{2,}[a-z]$", re.I)
     RX_UUID = re.compile (r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", re.I)
-    def _validate_param (self, params, required, ints, floats, emails, uuids, nones, **kargs):
+    def _validate_param (self, params, required, ints, floats, emails, uuids, nones, lists, strings, booleans, dicts, **kargs):
         params = params or {}
+
         if required:
             for each in required:
                 if not params.get (each):
                     return 'parameter {} required'.format (each)
+
         if ints:
             for each in ints:
+                try: val = params [each]
+                except KeyError: continue
+                if not val: val = 0
                 try:
-                    params [each] = int (params [each])
-                except KeyError:
-                    pass
+                    params [each] = int (val)
                 except ValueError:
-                    if not params [each]:
-                        params [each] = 0
-                    else:
-                        return 'parameter {} should be integer'.format (each)
+                    return 'parameter {} should be integer'.format (each)
         if floats:
             for each in floats:
+                try: val = params [each]
+                except KeyError: continue
+                if not val: val = 0.
                 try:
-                    params [each] = float (params [each])
-                except KeyError:
-                    pass
+                    params [each] = float (val)
                 except ValueError:
-                    if not params [each]:
-                        params [each] = 0.0
+                    return 'parameter {} should be float'.format (each)
+
+        if lists:
+            for each in lists:
+                try: val = params [each]
+                except KeyError: continue
+                if val is None: val = []
+                if not isinstance (val, (str, list, tuple)):
+                    return 'parameter {} should be list or comma/bar delimtered string'.format (each)
+                if isinstance (val, str):
+                    if val.find ("|") != -1:
+                        val = val.split ("|")
                     else:
-                        return 'parameter {} should be float'.format (each)
+                        val = val.split (",")
+                params [each] = val
+
+        if strings:
+            for each in strings:
+                try: val = params [each]
+                except KeyError: continue
+                if val is None: val = ''
+                if isinstance (val, str):
+                    params [each] = val
+                else:
+                    return 'parameter {} should be string'.format (each)
+
+        if dicts:
+            for each in dicts:
+                try: val = params [each]
+                except KeyError: continue
+                if val is None: val = {}
+                if isinstance (val, dict):
+                    params [each] = val
+                else:
+                    return 'parameter {} should be object'.format (each)
+
+        if booleans:
+            for each in booleans:
+                try: val = params [each]
+                except KeyError: continue
+                if val in ('yes', 'true'): val = True
+                elif val in ('no', 'false'): val = False
+                if val in (True, False):
+                    params [each] = val
+                else:
+                    return 'parameter {} should be boolean'.format (each)
+
         if emails:
             for fd in emails:
                 kargs [fd] = self.RX_EMAIL
@@ -167,18 +216,13 @@ class Decorators:
 
             op = ops [-1]
             val = (isinstance (cond, (list, tuple)) and type (cond [0]) or type (cond)) (val)
-            if op == "neq":
-                if val == cond:
+            if op in ("eq", "exact"):
+                if val != cond:
                     return 'parameter {} should be {}'.format (fd, cond)
-            elif op == "in":
-                if val not in cond:
-                    return 'parameter {} should be one of {}'.format (fd, cond)
-            elif op == "notin":
-                if val in cond:
-                    return 'parameter {} should be not one of {}'.format (fd, cond)
-            elif op == "between":
-                if not (cond [0] <= val <= cond [1]):
-                    return 'parameter {} should be between {} ~ {}'.format (fd, cond [0], cond [1])
+            elif op == "neq":
+                if val == cond:
+                    return 'parameter {} should not be {}'.format (fd, cond)
+
             elif op == "lte":
                 if val > cond:
                     return 'parameter {} should less or equal than {}'.format (fd, cond)
@@ -191,6 +235,35 @@ class Decorators:
             elif op == "gt":
                 if val <= cond:
                     return 'parameter {} should greater than {}'.format (fd, cond)
+            elif op == "between":
+                if not (cond [0] <= val <= cond [1]):
+                    return 'parameter {} should be between {} ~ {}'.format (fd, cond [0], cond [1])
+            elif op == "in":
+                if val not in cond:
+                    return 'parameter {} should be one of {}'.format (fd, cond)
+            elif op == "notin":
+                if val in cond:
+                    return 'parameter {} should be not one of {}'.format (fd, cond)
+
+            elif op == "startswith":
+                if not val.startswith (cond):
+                    return 'parameter {} should be start with {}'.format (fd, cond)
+            elif op == "endswith":
+                if not val.endswith (cond):
+                    return 'parameter {} should be end with {}'.format (fd, cond)
+            elif op == "nstartswith":
+                if val.startswith (cond):
+                    return 'parameter {} should be not start with {}'.format (fd, cond)
+            elif op == "nendswith":
+                if val.endswith (cond):
+                    return 'parameter {} should be not end with {}'.format (fd, cond)
+            elif op == "contains":
+                if val.find (cond) == -1:
+                    return 'parameter {} should be contain {}'.format (fd, cond)
+            elif op == "ncontains":
+                if val.find (cond) != -1:
+                    return 'parameter {} should be not contain {}'.format (fd, cond)
+
             else:
                 raise ValueError ("Unknown operator: {}".format (op))
 
@@ -205,7 +278,7 @@ class Decorators:
     def get_parameter_requirements (self, func_id):
         return self._parameter_caches.get (func_id, {})
 
-    def require (self, scope, required = None, ints = None, floats = None, uuids = None, emails = None, nones = None, **kargs):
+    def require (self, scope = 'ARGS', required = None, ints = None, floats = None, uuids = None, emails = None, nones = None, lists = None, strings = None, booleans = None, dicts = None, **kargs):
         def decorator(f):
             self.save_function_spec (f)
             func_id = self.get_func_id (f)
@@ -215,7 +288,7 @@ class Decorators:
             @wraps(f)
             def wrapper (was, *args, **kwargs):
                 scope_ = scope
-                if scope in ("FORM", "JSON"):
+                if scope in ("FORM", "JSON", "DATA"):
                     if was.request.method not in {"POST", "PUT", "PATCH"}:
                         return f (was, *args, **kwargs)
                     if scope == "JSON" and not was.request.get_header ("content-type", '').startswith ("application/json"):
@@ -235,7 +308,7 @@ class Decorators:
                         return f (was, *args, **kwargs)
 
                 validatable_args = getattr (was.request, scope_)
-                more_info = self._validate_param (validatable_args, required, ints, floats, emails, uuids, nones, **kargs)
+                more_info = self._validate_param (validatable_args, required, ints, floats, emails, uuids, nones, lists, strings, booleans, dicts, **kargs)
                 if more_info:
                     return was.response.adaptive_error ("400 Bad Request", 'missing or bad parameterin {}'.format (scope_), 40050, more_info)
 
