@@ -180,7 +180,7 @@ class WAS (wsgiappservice.WAS):
     def csrf_token_input (self):
         return '<input type="hidden" name="{}" value="{}">'.format (self.CSRF_NAME, self.csrf_token)
 
-    def verify_csrf (self, keep = False):
+    def verify_csrf (self, keep = True):
         token = self.request.get_header ('x-csrf-token', self.request.args.get (self.CSRF_NAME))
         if not token:
             return False
